#include <cxxtest/TestSuite.h>
#include <libgeodecomp/communication/hpxserializationwrapper.h>
#include <libgeodecomp/misc/testcell.h>
#include <libgeodecomp/storage/displacedgrid.h>

/**
 * Test model for use with Boost.Serialization
 */
class MyComplicatedCell1
{
public:
    class API : public LibGeoDecomp::APITraits::HasBoostSerialization
    {};

    template<typename NEIGHBORHOOD>
    void update(const NEIGHBORHOOD& hood, int nanoStep)
    {
    }

    inline bool operator==(const MyComplicatedCell1& other)
    {
        return
            (x     == other.x) &&
            (cargo == other.cargo);
    }

    template<typename ARCHIVE>
    void serialize(ARCHIVE& archive, int version)
    {
        archive & x;
        archive & cargo;
    }

    int x;
    std::vector<int> cargo;
};

class MyDummyCell
{
public:
    class API :
        public LibGeoDecomp::APITraits::HasSoA
    {};

    explicit MyDummyCell(const int x = 0, const double y = 0, const char z = 0) :
        x(x),
        y(y),
        z(z)
    {}

    inline bool operator==(const MyDummyCell& other)
    {
        return
            (x == other.x) &&
            (y == other.y) &&
            (z == other.z);
    }

    int x;
    double y;
    char z;
};

LIBFLATARRAY_REGISTER_SOA(MyDummyCell, ((int)(x))((double)(y))((char)(z)) )

using namespace LibGeoDecomp;

namespace LibGeoDecomp {

class DisplacedGridTest : public CxxTest::TestSuite
{
public:

    void testBoundingBox()
    {
        CoordBox<2> rect(Coord<2>(10, 11), Coord<2>(12, 13));
        DisplacedGrid<int> grid(rect);

        TS_ASSERT_EQUALS(rect, grid.boundingBox());
        TS_ASSERT_EQUALS(12, grid.getDimensions().x());
        TS_ASSERT_EQUALS(13, grid.getDimensions().y());
    }

    void testRegionConstructor()
    {
        CoordBox<2> rect(Coord<2>(44, 77), Coord<2>(55, 66));
        Region<2> region;
        region << rect;

        DisplacedGrid<int> grid(region);
        TS_ASSERT_EQUALS(rect, grid.boundingBox());
        TS_ASSERT_EQUALS(55, grid.getDimensions().x());
        TS_ASSERT_EQUALS(66, grid.getDimensions().y());
    }

    void testRemapRegion()
    {
        Region<3> region;
        region << Streak<3>(Coord<3>(11, 12, 13), 14);

        DisplacedGrid<int, Topologies::Torus<3>::Topology> grid;
        Region<3> remappedRegion = grid.remapRegion(region);
        TS_ASSERT_EQUALS(region, remappedRegion);
    }

    void testBoundingRegion()
    {
        Coord<3> origin(10, 11, 12);
        Coord<3> dim(13, 14, 15);

        CoordBox<3> rect(origin, dim);
        DisplacedGrid<double, Topologies::Cube<3>::Topology> grid(rect);

        Region<3> expectedRegion;
        expectedRegion << rect;

        TS_ASSERT_EQUALS(expectedRegion, grid.boundingRegion());
    }

    void testSetOrigin()
    {
        DisplacedGrid<int> grid(CoordBox<2>(Coord<2>(10, 20), Coord<2>(2, 3)));

        for (int y = 0; y < 3; ++y) {
            for (int x = 0; x < 2; ++x) {
                grid[Coord<2>(x + 10, y + 20)] = x * 10 + y;
            }
        }

        grid.setOrigin(Coord<2>(0, 50));
        for (int y = 0; y < 3; ++y) {
            for (int x = 0; x < 2; ++x) {
                TS_ASSERT_EQUALS(grid[Coord<2>(x + 0, y + 50)],  x * 10 + y);
            }
        }
    }

    void testResize()
    {
        CoordBox<3> oldBox(Coord<3>(1, 1, 2), Coord<3>(3, 4, 2));
        CoordBox<3> newBox(Coord<3>(4, 4, 4), Coord<3>(3, 3, 3));

        DisplacedGrid<double, Topologies::Torus<3>::Topology> grid(oldBox);
        grid[Coord<3>(1, 1, 2)] = 47;
        grid[Coord<3>(0, 0, 1)] = 11;
        TS_ASSERT_EQUALS(47, grid[Coord<3>(1, 1, 2)]);
        TS_ASSERT_EQUALS(11, grid[Coord<3>(3, 4, 3)]);
        TS_ASSERT_EQUALS(oldBox, grid.boundingBox());

        grid.resize(newBox);
        grid[Coord<3>(4, 4, 4)] = 27;
        grid[Coord<3>(3, 3, 3)] = 1;
        TS_ASSERT_EQUALS(27, grid[Coord<3>(4, 4, 4)]);
        TS_ASSERT_EQUALS(1,  grid[Coord<3>(6, 6, 6)]);
        TS_ASSERT_EQUALS(newBox, grid.boundingBox());
    }

    void testGetSetManyCells()
    {
        Coord<2> origin(20, 15);
        Coord<2> dim(30, 10);
        Coord<2> end = origin + dim;
        DisplacedGrid<TestCell<2> > testGrid(CoordBox<2>(origin, dim));

        int num = 200;
        for (int y = origin.y(); y < end.y(); y++) {
            for (int x = origin.x(); x < end.x(); x++) {
                testGrid[Coord<2>(x, y)] =
                    TestCell<2>(Coord<2>(x, y), testGrid.getDimensions());
                testGrid[Coord<2>(x, y)].testValue =  num++;
            }
        }

        TestCell<2> cells[5];
        testGrid.get(Streak<2>(Coord<2>(21, 18), 26), cells);

        for (int i = 0; i < 5; ++i) {
            TS_ASSERT_EQUALS(cells[i], testGrid.get(Coord<2>(i + 21, 18)));
        }

        for (int i = 0; i < 5; ++i) {
            cells[i].testValue = i + 1234;
        }
        testGrid.set(Streak<2>(Coord<2>(21, 18), 26), cells);

        for (int i = 0; i < 5; ++i) {
            TS_ASSERT_EQUALS(cells[i], testGrid.get(Coord<2>(i + 21, 18)));
        }
    }

    void testTopologicalNormalizationWithTorus()
    {
        /**
         * This test should check whether it is possible to let a
         * displaced grid act on a 15x10 torus as if it would cover a
         * 8x6 field shifted by an offset of (-3, -2), as outlined
         * below. The code will access the fields marked with a
         * through d.
         *
         *   0123456789abcde
         * 9 XXXXX_______cXX
         * 8 XXXXX_______aXX
         * 7 _______________
         * 6 _______________
         * 5 _______________
         * 4 _______________
         * 3 XXXXX_______XXX
         * 2 XXXXX_______XXX
         * 1 XXXXX_______XXX
         * 0 bXXXX_______XXd
         *
         */

        CoordBox<2> box(Coord<2>(-3, -2), Coord<2>(8, 6));
        DisplacedGrid<int, Topologies::Torus<2>::Topology, true> grid(
            box,
            -2,
            -2,
            Coord<2>(15, 10));

        for (int y = -2; y < 4; ++y) {
            for (int x = -3; x < 5; ++x) {
                grid[Coord<2>(x, y)] = (y + 3) * 10 + (x + 3);
            }
        }

        TS_ASSERT_EQUALS(10, grid[Coord<2>(-3, -2)]);
        TS_ASSERT_EQUALS(33, grid[Coord<2>( 0,  0)]);
        TS_ASSERT_EQUALS(20, grid[Coord<2>(12, 9)]);
        TS_ASSERT_EQUALS(32, grid[Coord<2>(14, 0)]);

        TS_ASSERT_EQUALS(21, grid.getNeighborhood(Coord<2>(12, 9))[Coord<2>(1, 0)]);

        Region<2> region;
        region << CoordBox<2>(Coord<2>( 0, 0), Coord<2>(5, 4))
               << CoordBox<2>(Coord<2>(12, 0), Coord<2>(3, 4))
               << CoordBox<2>(Coord<2>( 0, 8), Coord<2>(5, 2))
               << CoordBox<2>(Coord<2>(12, 8), Coord<2>(3, 2));

        std::vector<int> actual(region.size());
        grid.saveRegion(&actual, region);

        std::vector<int> expected;
        for (int y = 2; y < 6; ++y) {
            for (int x = 3; x < 8; ++x) {
                expected << (y + 1) * 10 + (x + 0);
            }
            for (int x = 0; x < 3; ++x) {
                expected << (y + 1) * 10 + (x + 0);
            }
        }

        for (int y = 0; y < 2; ++y) {
            for (int x = 3; x < 8; ++x) {
                expected << (y + 1) * 10 + (x + 0);
            }
            for (int x = 0; x < 3; ++x) {
                expected << (y + 1) * 10 + (x + 0);
            }
        }

        TS_ASSERT_EQUALS(actual, expected);
    }

    void testLoadSaveMember()
    {
        // basic setup:
        Selector<MyDummyCell> xSelector(&MyDummyCell::x, "x");
        Selector<MyDummyCell> ySelector(&MyDummyCell::y, "y");
        Selector<MyDummyCell> zSelector(&MyDummyCell::z, "z");

        Coord<2> offset(100, 200);
        Coord<2> dim(40, 20);
        CoordBox<2> box(offset, dim);
        DisplacedGrid<MyDummyCell, Topologies::Cube<2>::Topology> grid(box);
        for (CoordBox<2>::Iterator i = box.begin(); i != box.end(); ++i) {
            grid[*i] = MyDummyCell(i->x(), i->y(), 13);
        }

        Region<2> region;
        region << Streak<2>(Coord<2>(100, 200), 110)
               << Streak<2>(Coord<2>(110, 210), 120)
               << Streak<2>(Coord<2>(130, 219), 140);

        std::vector<int   > xVector(region.size(), -1);
        std::vector<double> yVector(region.size(), -1);
        std::vector<char  > zVector(region.size(), -1);

        grid.saveMember(&xVector[0], MemoryLocation::HOST, xSelector, region);
        grid.saveMember(&yVector[0], MemoryLocation::HOST, ySelector, region);
        grid.saveMember(&zVector[0], MemoryLocation::HOST, zSelector, region);

        Region<2>::Iterator cursor = region.begin();

        // test whether grid data is accurately copied back:
        for (std::size_t i = 0; i < region.size(); ++i) {
            TS_ASSERT_EQUALS(xVector[i], cursor->x());
            TS_ASSERT_EQUALS(yVector[i], cursor->y());
            TS_ASSERT_EQUALS(zVector[i], 13);
            ++cursor;
        }

        // modify vectors and copy back to grid:
        for (std::size_t i = 0; i < region.size(); ++i) {
            xVector[i] = 1000 + i;
            yVector[i] = 2000 + i;
            zVector[i] = i;
        }

        grid.loadMember(&xVector[0], MemoryLocation::HOST, xSelector, region);
        grid.loadMember(&yVector[0], MemoryLocation::HOST, ySelector, region);
        grid.loadMember(&zVector[0], MemoryLocation::HOST, zSelector, region);

        int counter = 0;
        for (Region<2>::Iterator i = region.begin(); i != region.end(); ++i) {
            TS_ASSERT_EQUALS(grid[*i], MyDummyCell(1000 + counter, 2000 + counter, counter));
            ++counter;
        }
    }

    void testLoadSaveRegion()
    {
        Coord<2> origin(20, 15);
        Coord<2> dim(30, 10);
        Coord<2> end = origin + dim;
        DisplacedGrid<TestCell<2> > testGrid(CoordBox<2>(origin, dim));

        int num = 200;
        for (int y = origin.y(); y < end.y(); y++) {
            for (int x = origin.x(); x < end.x(); x++) {
                testGrid[Coord<2>(x, y)] =
                    TestCell<2>(Coord<2>(x, y), testGrid.getDimensions());
                testGrid[Coord<2>(x, y)].testValue =  num++;
            }
        }

        std::vector<TestCell<2> > buffer(70);
        Region<2> region;
        region << Streak<2>(Coord<2>(20, 15), 50)
               << Streak<2>(Coord<2>(21, 16), 31)
               << Streak<2>(Coord<2>(20, 24), 50);
        TS_ASSERT_EQUALS(buffer.size(), region.size());

        testGrid.saveRegion(&buffer, region);

        Region<2>::Iterator iter = region.begin();
        for (int i = 0; i < 70; ++i) {
            TestCell<2> actual = testGrid.get(*iter);
            TestCell<2> expected(*iter, testGrid.getDimensions());
            int expectedIndex = 200 + (*iter - origin).toIndex(dim);
            expected.testValue = expectedIndex;

            TS_ASSERT_EQUALS(actual, expected);
            ++iter;
        }

        // manipulate test data:
        for (int i = 0; i < 70; ++i) {
            buffer[i].pos = Coord<2>(-i, -10);
        }

        int index = 0;
        testGrid.loadRegion(buffer, region);
        for (Region<2>::Iterator i = region.begin(); i != region.end(); ++i) {
            Coord<2> actual = testGrid.get(*i).pos;
            Coord<2> expected = Coord<2>(index, -10);
            TS_ASSERT_EQUALS(actual, expected);

            --index;
        }
    }

    void testLoadSaveRegionWithOffset()
    {
        Coord<2> origin(20, 15);
        Coord<2> dim(40, 15);
        Coord<2> end = origin + dim;
        DisplacedGrid<TestCell<2> > testGrid(CoordBox<2>(origin, dim));

        int num = 200;
        for (int y = origin.y(); y < end.y(); y++) {
            for (int x = origin.x(); x < end.x(); x++) {
                testGrid[Coord<2>(x, y)] =
                    TestCell<2>(Coord<2>(x, y), testGrid.getDimensions());
                testGrid[Coord<2>(x, y)].testValue =  num++;
            }
        }

        std::vector<TestCell<2> > buffer(90);
        Region<2> region;
        region << Streak<2>(Coord<2>(20, 15), 60)
               << Streak<2>(Coord<2>(21, 16), 31)
               << Streak<2>(Coord<2>(20, 29), 60);
        TS_ASSERT_EQUALS(buffer.size(), region.size());

        Region<2> regionWithOffset59;
        regionWithOffset59 << Streak<2>(Coord<2>(25, 24), 65)
                           << Streak<2>(Coord<2>(26, 25), 36)
                           << Streak<2>(Coord<2>(25, 38), 65);
        TS_ASSERT_EQUALS(buffer.size(), regionWithOffset59.size());

        Region<2> regionWithOffset24;
        regionWithOffset24 << Streak<2>(Coord<2>(22, 19), 62)
                           << Streak<2>(Coord<2>(23, 20), 33)
                           << Streak<2>(Coord<2>(22, 33), 62);
        TS_ASSERT_EQUALS(buffer.size(), regionWithOffset24.size());

        testGrid.saveRegion(&buffer, regionWithOffset59, Coord<2>(-5, -9));

        Region<2>::Iterator iter = region.begin();
        for (int i = 0; i < 90; ++i) {
            TestCell<2> actual = testGrid.get(*iter);
            TestCell<2> expected(*iter, testGrid.getDimensions());
            int expectedIndex = 200 + (*iter - origin).toIndex(dim);
            expected.testValue = expectedIndex;

            TS_ASSERT_EQUALS(actual, expected);
            ++iter;
        }

        // manipulate test data:
        for (int i = 0; i < 90; ++i) {
            buffer[i].pos = Coord<2>(-i, -20);
        }

        int index = 0;
        testGrid.loadRegion(buffer, regionWithOffset24, Coord<2>(-2, -4));
        for (Region<2>::Iterator i = region.begin(); i != region.end(); ++i) {
            Coord<2> actual = testGrid.get(*i).pos;
            Coord<2> expected = Coord<2>(index, -20);
            TS_ASSERT_EQUALS(actual, expected);

            --index;
        }
    }

    void testLoadSaveRegionWithBoostSerialization()
    {
#ifdef LIBGEODECOMP_WITH_BOOST_SERIALIZATION
        typedef DisplacedGrid<MyComplicatedCell1> GridType;

        Coord<2> dim(30, 20);
        CoordBox<2> box(Coord<2>(200, 100), dim);

        GridType sendGrid(box);
        GridType recvGrid(box);

        for (CoordBox<2>::Iterator i = box.begin(); i != box.end(); ++i) {
            MyComplicatedCell1 cell;
            cell.cargo << i->x();
            cell.cargo << i->y();
            cell.x = i->x() * 100 + i->y();
            sendGrid.set(*i, cell);
        }

        Region<2> region;
        region << Streak<2>(Coord<2>(2201, 1101), 2229)
               << Streak<2>(Coord<2>(2201, 1102), 2202)
               << Streak<2>(Coord<2>(2228, 1102), 2229)
               << Streak<2>(Coord<2>(2201, 1103), 2202)
               << Streak<2>(Coord<2>(2228, 1103), 2229)
               << Streak<2>(Coord<2>(2201, 1104), 2202)
               << Streak<2>(Coord<2>(2228, 1104), 2229)
               << Streak<2>(Coord<2>(2201, 1105), 2202)
               << Streak<2>(Coord<2>(2228, 1105), 2229)
               << Streak<2>(Coord<2>(2201, 1106), 2202)
               << Streak<2>(Coord<2>(2228, 1106), 2229)
               << Streak<2>(Coord<2>(2201, 1107), 2202)
               << Streak<2>(Coord<2>(2228, 1107), 2229)
               << Streak<2>(Coord<2>(2201, 1108), 2202)
               << Streak<2>(Coord<2>(2228, 1108), 2229)
               << Streak<2>(Coord<2>(2201, 1109), 2202)
               << Streak<2>(Coord<2>(2228, 1109), 2229)
               << Streak<2>(Coord<2>(2201, 1110), 2202)
               << Streak<2>(Coord<2>(2228, 1110), 2229)
               << Streak<2>(Coord<2>(2201, 1111), 2202)
               << Streak<2>(Coord<2>(2228, 1111), 2229)
               << Streak<2>(Coord<2>(2201, 1112), 2202)
               << Streak<2>(Coord<2>(2228, 1112), 2229)
               << Streak<2>(Coord<2>(2201, 1113), 2202)
               << Streak<2>(Coord<2>(2228, 1113), 2229)
               << Streak<2>(Coord<2>(2201, 1114), 2202)
               << Streak<2>(Coord<2>(2228, 1114), 2229)
               << Streak<2>(Coord<2>(2201, 1115), 2202)
               << Streak<2>(Coord<2>(2228, 1115), 2229)
               << Streak<2>(Coord<2>(2201, 1116), 2202)
               << Streak<2>(Coord<2>(2228, 1116), 2229)
               << Streak<2>(Coord<2>(2201, 1117), 2229);
        TS_ASSERT_EQUALS(region.size(), 86);
        Coord<2> offset(-2000, -1000);

        std::vector<char> buffer;
        sendGrid.saveRegion(&buffer, region, offset);

        for (Region<2>::Iterator i = region.begin(); i != region.end(); ++i) {
            TS_ASSERT_DIFFERS(sendGrid[*i + offset], recvGrid[*i + offset]);
        }
        recvGrid.loadRegion(buffer, region, offset);
        for (Region<2>::Iterator i = region.begin(); i != region.end(); ++i) {
            TS_ASSERT_EQUALS(sendGrid[*i + offset], recvGrid[*i + offset]);
        }
#endif
    }
};

}
